# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2019-2020 Max-Planck-Society


import time
import numpy as np
import ducc0.healpix as ph

rng = np.random.default_rng(42)


def report(name, vlen, ntry, nside, isnest, perf):
    print(name, ": ", perf*1e-6, "MOps/s", sep="")


def random_ptg(vlen):
    res = np.empty((vlen, 2), dtype=np.float64)
    res[:, 0] = np.arccos((rng.random(vlen)-0.5)*2)
    res[:, 1] = rng.random(vlen)*2*np.pi
    return res


def random_pix(nside, vlen):
    return rng.integers(low=0, high=12*nside*nside-1, size=vlen, dtype=np.int64)


def dummy(vlen):
    _ = np.zeros(vlen, dtype=np.int64)


def genperf(func, fname, inp, vlen, ntry, nside, isnest):
    cnt = 0
    t = time.time()
    while cnt < ntry:
        func(inp)
        cnt += 1
    t = time.time()-t
    p = (vlen*ntry)/t
    report(fname, vlen, ntry, nside, isnest, p)


def perf_pix2ang(vlen, ntry, nside, isnest):
    inp = random_pix(nside, vlen)
    base = ph.Healpix_Base(nside, "NEST" if isnest else "RING")
    genperf(base.pix2ang, "pix2ang", inp, vlen, ntry, nside, isnest)


def perf_ang2pix(vlen, ntry, nside, isnest):
    inp = random_ptg(vlen)
    base = ph.Healpix_Base(nside, "NEST" if isnest else "RING")
    genperf(base.ang2pix, "ang2pix", inp, vlen, ntry, nside, isnest)


def perf_pix2vec(vlen, ntry, nside, isnest):
    inp = random_pix(nside, vlen)
    base = ph.Healpix_Base(nside, "NEST" if isnest else "RING")
    genperf(base.pix2vec, "pix2vec", inp, vlen, ntry, nside, isnest)


def perf_vec2pix(vlen, ntry, nside, isnest):
    inp = ph.ang2vec(random_ptg(vlen))
    base = ph.Healpix_Base(nside, "NEST" if isnest else "RING")
    genperf(base.vec2pix, "vec2pix", inp, vlen, ntry, nside, isnest)


def perf_ring2nest(vlen, ntry, nside, isnest):
    inp = random_pix(nside, vlen)
    base = ph.Healpix_Base(nside, "NEST" if isnest else "RING")
    genperf(base.ring2nest, "ring2nest", inp, vlen, ntry, nside, isnest)


def perf_nest2ring(vlen, ntry, nside, isnest):
    inp = random_pix(nside, vlen)
    base = ph.Healpix_Base(nside, "NEST" if isnest else "RING")
    genperf(base.nest2ring, "nest2ring", inp, vlen, ntry, nside, isnest)


def perf_neighbors(vlen, ntry, nside, isnest):
    inp = random_pix(nside, vlen)
    base = ph.Healpix_Base(nside, "NEST" if isnest else "RING")
    genperf(base.neighbors, "neighbors", inp, vlen, ntry, nside, isnest)


def suite(vlen, ntry, nside, isnest):
    print("vlen=", vlen, ", ", "NEST" if isnest else "RING", sep="")
    dummy(vlen)
    perf_pix2ang(vlen, ntry, nside, isnest)
    perf_ang2pix(vlen, ntry, nside, isnest)
    perf_pix2vec(vlen, ntry, nside, isnest)
    perf_vec2pix(vlen, ntry, nside, isnest)
    perf_neighbors(vlen, ntry, nside, isnest)


nside = 512
ntry = 1000
print("nside=", nside, sep="")
for vlen in (1, 10, 100, 1000, 10000):
    for isnest in (True, False):
        suite(vlen, ntry, nside, isnest)
        perf_ring2nest(vlen, ntry, nside, isnest)
        perf_nest2ring(vlen, ntry, nside, isnest)
        print()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022-2025 Max-Planck-Society

import ducc0
import numpy as np
from time import time

try:
    import finufft
    have_finufft = True
except ImportError:
    have_finufft = False

class Bench12:
    def __init__(self, shape, npoints):
        self._shape = shape
        ndim = len(shape)
        # We create the random values in single precision, just to make sure
        # we don't change them by truncating in case we run a single precision
        # transform.

        # random nonuniform coordinates in [-pi; pi]
        self._coord = (2*np.pi*np.random.uniform(size=(npoints,ndim)) - np.pi).astype(np.float32)
        # random nonuniform strengths with -0.5 < re, im < 0.5
        self._points = (np.random.uniform(size=npoints)-0.5
                 + 1j * np.random.uniform(size=npoints)-0.5j).astype(np.complex64)
        # random uniform grid values -0.5 < re, im < 0.5
        self._values = (np.random.uniform(size=shape)-0.5
                 + 1j * np.random.uniform(size=shape)-0.5j).astype(np.complex64)

        # Produce "ground truth", i.e. run NUFFT with the best available
        # precision
        print("computing reference results with high precision ...")
        eps = 1.0001*ducc0.nufft.bestEpsilon(ndim=ndim, singleprec=False)
        self._res_fiducial_1 = ducc0.nufft.nu2u(
            points=self._points.astype(np.complex128),
            coord=self._coord.astype(np.float64),
            forward=True,
            epsilon=eps,
            nthreads=0,
            verbosity=0,
            out=np.empty(shape, dtype=np.complex128))

        self._res_fiducial_2 = ducc0.nufft.u2nu(
            grid=self._values.astype(np.complex128),
            coord=self._coord.astype(np.float64),
            forward=True,
            epsilon=eps,
            nthreads=0,
            verbosity=0)
        print("done")

    def run(self, epsilon, singleprec, nthreads):
        rdtype = np.float32 if singleprec else np.float64
        dtype = np.complex64 if singleprec else np.complex128

        res={}
        res["shape"] = self._shape
        res["npoints"] = self._coord.shape[0]
        res["epsilon"] = epsilon
        res["nthreads"] = nthreads
        res["singleprec"] = singleprec

        shape = self._shape
        ndim = len(shape)
        npoints = self._coord.shape[0]
        coord = self._coord.astype(rdtype)
        points = self._points.astype(dtype)
        values = self._values.astype(dtype)

        out = np.ones(shape, dtype=dtype)
        t0 = time()
        plan = ducc0.nufft.plan(nu2u=True, coord=coord, grid_shape=shape, epsilon=epsilon, nthreads=nthreads)
        res["ducc_1_planned_time_plan"] = time()-t0
        t0 = time()
        out = plan.nu2u(points=points, forward=True, verbosity=0, out=out)
        res["ducc_1_planned_time_exec"] = time()-t0
        res["ducc_1_planned_err"] = ducc0.misc.l2error(out, self._res_fiducial_1)
        print(f"ducc0,     planned, type 1: time={res["ducc_1_planned_time_exec"]}, L2 error={res["ducc_1_planned_err"]}")
        del plan, out

        out=np.ones(shape=(npoints,), dtype=dtype)
        t0 = time()
        plan = ducc0.nufft.plan(nu2u=False, coord=coord, grid_shape=shape, epsilon=epsilon, nthreads=nthreads)
        res["ducc_2_planned_time_plan"] = time()-t0
        t0 = time()
        out = plan.u2nu(grid=values, forward=True, verbosity=0, out=out)
        res["ducc_2_planned_time_exec"] = time()-t0
        res["ducc_2_planned_err"] = ducc0.misc.l2error(out, self._res_fiducial_2)
        print(f"ducc0,     planned, type 2: time={res["ducc_2_planned_time_exec"]}, L2 error={res["ducc_2_planned_err"]}")
        del plan, out

        out = np.ones(shape, dtype=dtype)
        t0 = time()
        out = ducc0.nufft.nu2u(points=points, coord=coord, forward=True, epsilon=epsilon, nthreads=nthreads, verbosity=0, out=out)
        res["ducc_1_unplanned_time_full"] = time()-t0
        res["ducc_1_unplanned_err"] = ducc0.misc.l2error(out, self._res_fiducial_1)
        print(f"ducc0,   unplanned, type 1: time={res["ducc_1_unplanned_time_full"]}, L2 error={res["ducc_1_unplanned_err"]}")
        del out

        out=np.ones(shape=(npoints,), dtype=dtype)
        t0 = time()
        out = ducc0.nufft.u2nu(grid=values, coord=coord, forward=True, epsilon=epsilon, nthreads=nthreads, verbosity=0, out=out)
        res["ducc_2_unplanned_time_full"] = time()-t0
        res["ducc_2_unplanned_err"] = ducc0.misc.l2error(out, self._res_fiducial_2)
        print(f"ducc0,   unplanned, type 2: time={res["ducc_2_unplanned_time_full"]}, L2 error={res["ducc_2_unplanned_err"]}")
        del out

        if not have_finufft:
            return res

        import finufft
        coord = tuple(np.ascontiguousarray(coord[:,i]) for i in range(coord.shape[1]))
        func1=[finufft.nufft1d1, finufft.nufft2d1, finufft.nufft3d1]
        func2=[finufft.nufft1d2, finufft.nufft2d2, finufft.nufft3d2]

        t0 = time()
        plan = finufft.Plan(1, self._shape, 1, eps=epsilon, isign=-1,
                            dtype="complex64" if singleprec else "complex128",
                            nthreads=nthreads, debug=0)
        plan.setpts(*coord)
        res["finufft_1_planned_time_plan"] = time()-t0
        out = np.ones(shape, dtype=dtype)
        t0 = time()
        out = plan.execute(points, out=out)
        res["finufft_1_planned_time_exec"] = time()-t0
        res["finufft_1_planned_err"] = ducc0.misc.l2error(out, self._res_fiducial_1)
        print(f"Finufft,   planned, type 1: time={res["finufft_1_planned_time_exec"]}, L2 error={res["finufft_1_planned_err"]}")
        del plan, out

        t0 = time()
        plan = finufft.Plan(2, shape, 1, eps=epsilon, isign=-1,
                            dtype="complex64" if singleprec else "complex128",
                            nthreads=nthreads, debug=0)
        plan.setpts(*coord)
        res["finufft_2_planned_time_plan"] = time()-t0
        out = np.ones((npoints,), dtype=dtype)
        t0 = time()
        out = plan.execute(values, out=out)
        res["finufft_2_planned_time_exec"] = time()-t0
        res["finufft_2_planned_err"] = ducc0.misc.l2error(out, self._res_fiducial_2)
        print(f"Finufft,   planned, type 2: time={res["finufft_2_planned_time_exec"]}, L2 error={res["finufft_2_planned_err"]}")
        del plan, out

        out = np.ones(shape, dtype=dtype)
        t0=time()
        out = func1[ndim-1](*coord, points, out=out, eps=epsilon, isign=-1, nthreads=nthreads, debug=0)
        res["finufft_1_unplanned_time_exec"] = time()-t0
        res["finufft_1_unplanned_err"] = ducc0.misc.l2error(out, self._res_fiducial_1)
        print(f"Finufft, unplanned, type 1: time={res["finufft_1_unplanned_time_exec"]}, L2 error={res["finufft_1_unplanned_err"]}")
        del out

        out = np.ones((npoints,), dtype=dtype)
        t0=time()
        out = func2[ndim-1](*coord, values, out=out, eps=epsilon, isign=-1, nthreads=nthreads, debug=0)
        res["finufft_2_unplanned_time_exec"] = time()-t0
        res["finufft_2_unplanned_err"] = ducc0.misc.l2error(out, self._res_fiducial_2)
        print(f"Finufft, unplanned, type 2: time={res["finufft_2_unplanned_time_exec"]}, L2 error={res["finufft_2_unplanned_err"]}")
        del out

        return res


class Bench3:
    def __init__(self, npoints_in, npoints_out, minmax_in, minmax_out):
        self._npoints_in = npoints_in
        self._npoints_out = npoints_out
        self._minmax_in = minmax_in
        self._minmax_out = minmax_out
        ndim = minmax_in.shape[1]
        # We create the random values in single precision, just to make sure
        # we don't change them by truncating in case we run a single precision
        # transform.

        # random nonuniform input coordinates
        self._coord_in = np.random.uniform(size=(npoints_in,ndim)).astype(np.float32)
        self._coord_in *= minmax_in[1]-minmax_in[0]
        self._coord_in += minmax_in[0]
        # random nonuniform input coordinates
        self._coord_out = np.random.uniform(size=(npoints_out,ndim)).astype(np.float32)
        self._coord_out *= minmax_out[1]-minmax_out[0]
        self._coord_out += minmax_out[0]
        # random nonuniform strengths with -0.5 < re, im < 0.5
        self._points = (np.random.uniform(size=npoints_in)-0.5
                 + 1j * np.random.uniform(size=npoints_in)-0.5j).astype(np.complex64)

        # Produce "ground truth", i.e. run NUFFT with the best available
        # precision
        print("computing reference results with high precision ...")
        eps = 1.0001*ducc0.nufft.bestEpsilon(ndim=ndim, singleprec=False)
        self._res_fiducial = ducc0.nufft.experimental.nu2nu(
            points_in=self._points.astype(np.complex128),
            coord_in=self._coord_in.astype(np.float64),
            coord_out=self._coord_out.astype(np.float64),
            forward=True, epsilon=eps, verbosity=0, nthreads=0)
        print("done")

    def run(self, epsilon, singleprec, nthreads):
        rdtype = np.float32 if singleprec else np.float64
        dtype = np.complex64 if singleprec else np.complex128

        res={}
        res["npoints_in"] = self._coord_in.shape[0]
        res["npoints_out"] = self._coord_out.shape[0]
        res["epsilon"] = epsilon
        res["nthreads"] = nthreads
        res["singleprec"] = singleprec

        ndim = self._minmax_in.shape[1]
        npoints_in = self._coord_in.shape[0]
        coord_in = self._coord_in.astype(rdtype)
        npoints_out = self._coord_out.shape[0]
        coord_out = self._coord_out.astype(rdtype)
        points = self._points.astype(dtype)

        out = np.ones((npoints_out,), dtype=dtype)
        t0 = time()
        plan = ducc0.nufft.experimental.plan3(coord_in=coord_in,
           coord_out=coord_out, epsilon=epsilon, verbosity=0, nthreads=nthreads)
        res["ducc_3_planned_time_plan"] = time()-t0
        t0 = time()
        out = plan.exec(forward=True,points_in=points,points_out=out)
        res["ducc_3_planned_time_exec"] = time()-t0
        res["ducc_3_planned_err"] = ducc0.misc.l2error(out, self._res_fiducial)
        print(f"ducc0,     planned, type 3: time={res["ducc_3_planned_time_exec"]}, L2 error={res["ducc_3_planned_err"]}")
        del plan, out

        out = np.ones((npoints_in,), dtype=dtype)
        t0 = time()
        out = ducc0.nufft.experimental.nu2nu(points_in=points, coord_in=coord_in, coord_out=coord_out, forward=True, epsilon=epsilon, verbosity=0, nthreads=nthreads, points_out=out)
        res["ducc_3_unplanned_time_full"] = time()-t0
        res["ducc_3_unplanned_err"] = ducc0.misc.l2error(out, self._res_fiducial)
        print(f"ducc0,   unplanned, type 3: time={res["ducc_3_unplanned_time_full"]}, L2 error={res["ducc_3_unplanned_err"]}")
        del out

        if not have_finufft:
            return res

        import finufft
        coord_in = tuple(np.ascontiguousarray(coord_in[:,i]) for i in range(coord_in.shape[1]))
        coord_out = tuple(np.ascontiguousarray(coord_out[:,i]) for i in range(coord_out.shape[1]))

        out = np.ones((npoints_out,), dtype=dtype)
        plan = finufft.Plan(3, ndim, eps=epsilon, isign=-1, dtype=dtype, nthreads=nthreads)
        args = list(coord_in) + [None]*(3-ndim) + list(coord_out)
        plan.setpts(*args)
        t0 = time()
        out = plan.execute(points, out=out)
        res["finufft_3_planned_time_exec"] = time()-t0
        res["finufft_3_planned_err"] = ducc0.misc.l2error(out, self._res_fiducial)
        print(f"finufft,   planned, type 3: time={res["finufft_3_planned_time_exec"]}, L2 error={res["finufft_3_planned_err"]}")
        del plan, out


        out = np.ones((npoints_out,), dtype=dtype)
        func=[finufft.nufft1d3, finufft.nufft2d3, finufft.nufft3d3]
        t0 = time()
        out = func[ndim-1](*coord_in, points, *coord_out, out=out, eps=epsilon, isign=-1, nthreads=nthreads)
        res["finufft_3_unplanned_time_full"] = time()-t0
        res["finufft_3_unplanned_err"] = ducc0.misc.l2error(out, self._res_fiducial)
        print(f"finufft, unplanned, type 3: time={res["finufft_3_unplanned_time_full"]}, L2 error={res["finufft_3_unplanned_err"]}")
        del out

        return res


def plot(res, fname):
    import matplotlib.pyplot as plt
    fct = 1e9/res[0]["npoints"]
    have_finufft = "finufft_1_planned_time_exec" in res[0]
    tducc1 = fct*np.array([r["ducc_1_unplanned_time_full"] for r in res])
    tducc2 = fct*np.array([r["ducc_2_unplanned_time_full"] for r in res])
    tducct1 = fct*np.array([r["ducc_1_planned_time_exec"] for r in res])
    tducct2 = fct*np.array([r["ducc_2_planned_time_exec"] for r in res])
    educc1 = np.array([r["ducc_1_unplanned_err"] for r in res])
    educc2 = np.array([r["ducc_2_unplanned_err"] for r in res])
    educct1 = np.array([r["ducc_1_planned_err"] for r in res])
    educct2 = np.array([r["ducc_2_planned_err"] for r in res])
    if have_finufft:
        tfinufftt1 = fct*np.array([r["finufft_1_planned_time_exec"] for r in res])
        tfinufftt2 = fct*np.array([r["finufft_2_planned_time_exec"] for r in res])
        tfinufft1 = fct*np.array([r["finufft_1_unplanned_time_exec"] for r in res])
        tfinufft2 = fct*np.array([r["finufft_2_unplanned_time_exec"] for r in res])
        efinufftt1 = np.array([r["finufft_1_planned_err"] for r in res])
        efinufftt2 = np.array([r["finufft_2_planned_err"] for r in res])
        efinufft1 = np.array([r["finufft_1_unplanned_err"] for r in res])
        efinufft2 = np.array([r["finufft_2_unplanned_err"] for r in res])
    eps = np.array([r["epsilon"] for r in res])
    plt.xscale("log")
    plt.yscale("log")
    plt.plot(educc1,tducc1,label="ducc unplanned, type 1")
    plt.plot(educc2,tducc2,label="ducc unplanned, type 2")
    plt.plot(educct1,tducct1,label="ducc planned, type 1")
    plt.plot(educct2,tducct2,label="ducc planned, type 2")
    if have_finufft:
        plt.plot(efinufftt1,tfinufftt1,label="finufft planned, type 1")
        plt.plot(efinufftt2,tfinufftt2,label="finufft planned, type 2")
        plt.plot(efinufft1,tfinufft1,label="finufft unplanned, type 1")
        plt.plot(efinufft2,tfinufft2,label="finufft unplanned, type 2")
    plt.title("shape={}, npoints={}, nthreads={}".format(res[0]["shape"], res[0]["npoints"], res[0]["nthreads"]))
    plt.xlabel("real error")
    plt.ylabel("ns per nonuniform point")
    plt.legend()
    plt.savefig(fname, bbox_inches='tight')
    plt.close()


def plot3(res, fname):
    import matplotlib.pyplot as plt
    fct = 1e9/(res[0]["npoints_in"]+ res[0]["npoints_out"])
    have_finufft = "finufft_3_planned_time_exec" in res[0]
    tducc3 = fct*np.array([r["ducc_3_unplanned_time_full"] for r in res])
    tducct3 = fct*np.array([r["ducc_3_planned_time_exec"] for r in res])
    educc3 = np.array([r["ducc_3_unplanned_err"] for r in res])
    educct3 = np.array([r["ducc_3_planned_err"] for r in res])
    if have_finufft:
        tfinufftt3 = fct*np.array([r["finufft_3_planned_time_exec"] for r in res])
        tfinufft3 = fct*np.array([r["finufft_3_unplanned_time_full"] for r in res])
        efinufftt3 = np.array([r["finufft_3_planned_err"] for r in res])
        efinufft3 = np.array([r["finufft_3_unplanned_err"] for r in res])
    eps = np.array([r["epsilon"] for r in res])
    plt.xscale("log")
    plt.yscale("log")
    plt.plot(educc3,tducc3,label="ducc unplanned, type 3")
    plt.plot(educct3,tducct3,label="ducc planned, type 3")
    if have_finufft:
        plt.plot(efinufft3,tfinufft3,label="finufft unplanned, type 3")
        plt.plot(efinufftt3,tfinufftt3,label="finufft planned, type 3")
    plt.title("npoints_in={}, npoint_out={}, nthreads={}".format(res[0]["npoints_in"], res[0]["npoints_out"], res[0]["nthreads"]))
    plt.xlabel("real error")
    plt.ylabel("ns per nonuniform point")
    plt.legend()
    plt.savefig(fname, bbox_inches='tight')
    plt.close()


def runbench12(shape, npoints, nthreads, fname, singleprec=False):
    res=[]
    mybench = Bench12(shape, npoints)
    if singleprec:
        epslist = [[2.5e-7, 4.5e-7, 8.2e-7][len(shape)-1], 1e-5, 1e-4, 1e-3, 1e-2]
    else:
        epslist = [[4e-15, 8e-15, 2e-14][len(shape)-1], 1e-13, 1e-12, 1e-11, 1e-10, 1e-9, 1e-8, 1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2]
    for eps in epslist:
        print(f"N={shape}, M={npoints}, epsilon={eps}, nthreads={nthreads}:")
        res.append(mybench.run(eps, singleprec, nthreads))
        print()
    plot(res, fname)

def runbench3(npoints_in, npoints_out, minmax_in, minmax_out, nthreads, fname, singleprec=False):
    res=[]
    mybench = Bench3(npoints_in, npoints_out, minmax_in, minmax_out)
    ndim = minmax_in.shape[1]
    if singleprec:
        epslist = [[2.5e-7, 4.5e-7, 8.2e-7][ndim-1], 1e-5, 1e-4, 1e-3, 1e-2]
    else:
        epslist = [[4e-15, 8e-15, 2e-14][ndim-1], 1e-13, 1e-12, 1e-11, 1e-10, 1e-9, 1e-8, 1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2]
    for eps in epslist:
        print(f"{ndim}D, M_in={npoints_in}, M_out={npoints_out}, epsilon={eps}, nthreads={nthreads}:")
        res.append(mybench.run(eps, singleprec, nthreads))
        print()
    plot3(res, fname)


singleprec = False
# FINUFFT benchmarks, analogous to figures 6.1-6.3 in the 2018 paper
if True:
    runbench12((   1000000,),  10000000, 1, "finufft_1d_serial.png"  , singleprec)
    runbench12(( 1000,1000,),  10000000, 1, "finufft_2d_serial.png"  , singleprec)
    runbench12((100,100,100),  10000000, 1, "finufft_3d_serial.png"  , singleprec)
    runbench12((  10000000,), 100000000, 8, "finufft_1d_parallel.png", singleprec)
    runbench12(( 3162,3162,), 100000000, 8, "finufft_2d_parallel.png", singleprec)
    runbench12((216,216,216), 100000000, 8, "finufft_3d_parallel.png", singleprec)
# NFFT.jl benchmarks, lower nonuniform point density
if True:
    runbench12(( 512*512,),  512*512, 1, "bench_1d.png", singleprec)
    runbench12(( 512,512,),  512*512, 1, "bench_2d.png", singleprec)
    runbench12((64,64,64,), 64*64*64, 1, "bench_3d.png", singleprec)
# some preliminary type 3 benchmarks
if True:
    # helper function to create coordinate ranges fpr NU points
    def make_ranges(xmin, xmax, ymin=None, ymax=None, zmin=None, zmax=None):
        if ymin is None:
            return np.array([[xmin], [xmax]])
        if zmin is None:
            return np.array([[xmin, ymin], [xmax, ymax]])
        return np.array([[xmin, ymin, zmin], [xmax, ymax, zmax]])
    minmax = make_ranges(-1., 1.)
    runbench3(1000000,  1000000, minmax, minmax, 1, "finufft_1d_type3_serial.png", singleprec)
    runbench3(10000000,  10000000, minmax, minmax, 8, "finufft_1d_type3_parallel.png"  , singleprec)
    minmax = make_ranges(-1., 1., -1., 1.)
    runbench3(1000000,  1000000, minmax, minmax, 1, "finufft_2d_type3_serial.png", singleprec)
    runbench3(10000000,  10000000, minmax, minmax, 8, "finufft_2d_type3_parallel.png"  , singleprec)
    minmax = make_ranges(-1., 1., -1., 1., -1., 1.)
    runbench3(1000000,  1000000, minmax, minmax, 1, "finufft_3d_type3_serial.png", singleprec)
    runbench3(10000000,  10000000, minmax, minmax, 8, "finufft_3d_type3_parallel.png"  , singleprec)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020-2025 Max-Planck-Society

from itertools import product

import ducc0.wgridder as ng
import ducc0
from ducc0.misc import vdot
try:
    import finufft
    have_finufft = True
except ImportError:
    have_finufft = False
import numpy as np
import pytest
from numpy.testing import assert_allclose

pmp = pytest.mark.parametrize
SPEEDOFLIGHT = 299792458.


def explicit_gridder(uvw, freq, ms, wgt, nxdirty, nydirty, xpixsize, ypixsize,
                     apply_w, mask, divide_by_n=True):
    x, y = np.meshgrid(*[-ss/2 + np.arange(ss) for ss in [nxdirty, nydirty]],
                       indexing='ij')
    x *= xpixsize
    y *= ypixsize
    res = np.zeros((nxdirty, nydirty))
    eps = x**2+y**2
    if apply_w:
        tmp = np.sqrt(np.abs(1.-eps))
        nm1 = np.where(eps<=1, tmp-1., -tmp-1.)
        n = nm1+1
    else:
        nm1 = 0.
        n = 1.
    for row in range(ms.shape[0]):
        for chan in range(ms.shape[1]):
            if mask is not None and mask[row, chan] == 0:
                continue
            phase = ((2j*np.pi*freq[chan]/SPEEDOFLIGHT) *
                     (x*uvw[row, 0] + y*uvw[row, 1] - uvw[row, 2]*nm1))
            if wgt is None:
                res += (ms[row, chan]*np.exp(phase)).real
            else:
                res += ((ms[row, chan]*wgt[row, chan])
                        * np.exp(phase)).real
    return np.where(n>0, res/n, 0.) if divide_by_n else res

def explicit_degridder(uvw, freq, dirty, wgt, xpixsize, ypixsize,
                       apply_w, mask, divide_by_n=True):
    nxdirty, nydirty = dirty.shape
    x, y = np.meshgrid(*[-ss/2 + np.arange(ss) for ss in [nxdirty, nydirty]],
                       indexing='ij')
    x *= xpixsize
    y *= ypixsize
    res = np.zeros((uvw.shape[0], freq.shape[0]), dtype=np.complex128)
    eps = x**2+y**2
    if apply_w:
        tmp = np.sqrt(np.abs(1.-eps))
        nm1 = np.where(eps<=1, tmp-1., -tmp-1.)
        n = nm1+1
    else:
        nm1 = 0.
        n = 1.
    dirty2 = np.where(n>0, dirty/n, 0.) if divide_by_n else dirty
    for row in range(uvw.shape[0]):
        for chan in range(freq.shape[0]):
            if mask is not None and mask[row, chan] == 0:
                continue
            phase = ((-2j*np.pi*freq[chan]/SPEEDOFLIGHT) *
                     (x*uvw[row, 0] + y*uvw[row, 1] - uvw[row, 2]*nm1))
            if wgt is None:
                res[row, chan] = np.sum(dirty2*np.exp(phase))
            else:
                res[row, chan] = np.sum((dirty2*wgt[row, chan]
                        * np.exp(phase)))
    return res


def with_finufft(uvw, freq, ms, wgt, nxdirty, nydirty, xpixsize, ypixsize,
                 mask, epsilon):
    u = np.fmod(np.outer(uvw[:, 0], freq)*(xpixsize/SPEEDOFLIGHT), 1.)*2*np.pi
    v = np.fmod(np.outer(uvw[:, 1], freq)*(ypixsize/SPEEDOFLIGHT), 1.)*2*np.pi
    if wgt is not None:
        ms = ms*wgt
    if mask is not None:
        ms = ms*mask
    eps = epsilon/10  # Apparently finufft measures epsilon differently
    # Plan on the fly
    res0 = finufft.nufft2d1(u.ravel(), v.ravel(), ms.ravel(), (nxdirty,
                            nydirty), eps=eps).real
    # Plan beforehand
    plan = finufft.Plan(1, (nxdirty, nydirty), eps=eps)
    plan.setpts(u.ravel(), v.ravel())
    res1 = plan.execute(ms.ravel()).real
    np.testing.assert_allclose(res0, res1)
    return res0


def vis2dirty_with_faceting(*, nfacets_x=1, nfacets_y=1, npix_x, npix_y,
                            center_x=0., center_y=0., dirty=None, **kwargs):
    import ducc0.wgridder as wgridder
    if dirty is None:
        dtype = np.float32 if kwargs["vis"].dtype == np.complex64 else np.float64
        dirty = np.zeros((npix_x, npix_y), dtype)
    else:
        if npix_x != 0 and npix_x != dirty.shape[0]:
            raise ValueError("bad npix_x")
        if npix_y != 0 and npix_y != dirty.shape[1]:
            raise ValueError("bad npix_y")
        npix_x, npix_y = dirty.shape

    istep = (npix_x+nfacets_x-1) // nfacets_x
    istep += istep % 2  # make even
    jstep = (npix_y+nfacets_y-1) // nfacets_y
    jstep += jstep % 2  # make even

    tdirty = None
    for i in range(nfacets_x):
        cx = center_x + kwargs["pixsize_x"]*0.5*((2*i+1)*istep-npix_x)
        imax = min((i+1)*istep, dirty.shape[0])
        for j in range(nfacets_y):
            cy = center_y + kwargs["pixsize_y"]*0.5*((2*j+1)*jstep-npix_y)
            jmax = min((j+1)*jstep, dirty.shape[1])
            if imax == (i+1)*istep and jmax == (j+1)*jstep:
                tmp = np.ascontiguousarray(dirty[i*istep:imax, j*jstep:jmax])  # TEMPORARY until fixed in sycl-wgridder
                _ = wgridder.vis2dirty(**kwargs,
                                       center_x=cx, center_y=cy,
                                       npix_x=istep, npix_y=jstep,
                                       dirty=tmp)
                dirty[i*istep:imax, j*jstep:jmax] = tmp  # TEMPORARY until fixed in sycl-wgridder
            else:
                tdirty = wgridder.vis2dirty(
                    **kwargs, center_x=cx, center_y=cy,
                    npix_x=istep, npix_y=jstep, dirty=tdirty)
                dirty[i*istep:imax, j*jstep:jmax] = tdirty[:imax-i*istep, :jmax-j*jstep]
    return dirty


def dirty2vis_with_faceting(*, nfacets_x=1, nfacets_y=1, center_x=0., center_y=0.,
                            dirty, vis=None, **kwargs):
    import ducc0.wgridder as wgridder
    npix_x, npix_y = dirty.shape
    istep = (npix_x+nfacets_x-1) // nfacets_x
    istep += istep % 2  # make even
    jstep = (npix_y+nfacets_y-1) // nfacets_y
    jstep += jstep % 2  # make even

    tvis = None
    for i in range(nfacets_x):
        cx = center_x + kwargs["pixsize_x"]*0.5*((2*i+1)*istep-npix_x)
        imax = min((i+1)*istep, dirty.shape[0])
        for j in range(nfacets_y):
            cy = center_y + kwargs["pixsize_y"]*0.5*((2*j+1)*jstep-npix_y)
            jmax = min((j+1)*jstep, dirty.shape[1])
            if imax == (i+1)*istep and jmax == (j+1)*jstep:
                tdirty = dirty[i*istep:imax, j*jstep:jmax]
                tdirty = np.ascontiguousarray(tdirty)  # TEMPORARY until fixed in sycl-wgridder
            else:
                tdirty = np.zeros((istep, jstep), dtype=dirty.dtype)
                tdirty[:imax-i*istep, :jmax-j*jstep] = dirty[i*istep:imax, j*jstep:jmax]

            if i == 0 and j == 0:
                vis = wgridder.dirty2vis(**kwargs, dirty=tdirty,
                                         center_x=cx, center_y=cy, vis=vis)
            else:
                tvis = wgridder.dirty2vis(**kwargs, dirty=tdirty,
                                          center_x=cx, center_y=cy, vis=tvis)
                vis += tvis
    return vis


@pmp('nx', [(2, 2), (32, 3), (128, 2)])
@pmp('ny', [(2, 2), (128, 2), (250, 5)])
@pmp("nrow", (1, 2, 27))
@pmp("nchan", (1, 5))
@pmp("epsilon", (1e-1, 3e-5, 2e-13))
@pmp("singleprec", (True, False))
@pmp("wstacking", (True, False))
@pmp("use_wgt", (True, False))
@pmp("use_mask", (False, True))
@pmp("nthreads", (1, 2))
def test_adjointness_ms2dirty(nx, ny, nrow, nchan, epsilon,
                              singleprec, wstacking, use_wgt, nthreads,
                              use_mask):
    (nxdirty, nxfacets), (nydirty, nyfacets) = nx, ny
    if singleprec and epsilon < 1e-6:
        pytest.skip()
    rng = np.random.default_rng(43)
    pixsizex = np.pi/180/60/nxdirty*0.2398
    pixsizey = np.pi/180/60/nxdirty
    f0 = 1e9
    freq = f0 + np.arange(nchan)*(f0/nchan)
    uvw = (rng.random((nrow, 3))-0.5)/(pixsizey*f0/SPEEDOFLIGHT)
    ms = rng.random((nrow, nchan))-0.5 + 1j*(rng.random((nrow, nchan))-0.5)
    wgt = rng.uniform(0.9, 1.1, (nrow, nchan)) if use_wgt else None
    mask = (rng.uniform(0, 1, (nrow, nchan)) > 0.5).astype(np.uint8) \
        if use_mask else None
    dirty = rng.random((nxdirty, nydirty))-0.5
    nu = nv = 0
    if singleprec:
        ms = ms.astype("c8")
        dirty = dirty.astype("f4")
        if wgt is not None:
            wgt = wgt.astype("f4")

    def check(d2, m2):
        ref = max(vdot(ms, ms).real, vdot(m2, m2).real,
                  vdot(dirty, dirty).real, vdot(d2, d2).real)
        tol = 3e-5*ref if singleprec else 2e-13*ref
        assert_allclose(vdot(ms, m2).real, vdot(d2, dirty), rtol=3*tol)

    dirty2 = ng.ms2dirty(uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
                         pixsizey, nu, nv, epsilon, wstacking, nthreads, 0,
                         mask).astype("f8")
    ms2 = ng.dirty2ms(uvw, freq, dirty, wgt, pixsizex, pixsizey, nu, nv,
                      epsilon, wstacking, nthreads, 0, mask).astype("c16")
    check(dirty2, ms2)

    dirty2 = vis2dirty_with_faceting(nfacets_x=nxfacets, nfacets_y=nyfacets,
                                     uvw=uvw, freq=freq,
                                     vis=ms, wgt=wgt, npix_x=nxdirty,
                                     npix_y=nydirty, pixsize_x=pixsizex,
                                     pixsize_y=pixsizey, epsilon=epsilon,
                                     do_wgridding=wstacking, nthreads=nthreads,
                                     mask=mask).astype("f8")
    ms2 = dirty2vis_with_faceting(nfacets_x=nxfacets, nfacets_y=nyfacets,
                                  uvw=uvw, freq=freq,
                                  dirty=dirty, wgt=wgt, pixsize_x=pixsizex,
                                  pixsize_y=pixsizey, epsilon=epsilon,
                                  do_wgridding=wstacking, nthreads=nthreads,
                                  mask=mask).astype("c16")
    check(dirty2, ms2)


@pmp('nx', [(2, 2), (6, 2), (18, 2), (66, 4)])
@pmp('ny', [(2, 2), (64, 2)])
@pmp("nrow", (1, 2, 27))
@pmp("nchan", (1, 5))
@pmp("epsilon", (1e-2, 1e-4, 1e-7))
@pmp("singleprec", (False,))
@pmp("wstacking", (False, True))
@pmp("use_wgt", (True,))
@pmp("use_mask", (True,))
@pmp("nthreads", (1, 2))
@pmp("fov", (0.001, 0.1, 20.))
def test_ms2dirty_against_wdft2(nx, ny, nrow, nchan, epsilon,
                                singleprec, wstacking, use_wgt, use_mask, fov,
                                nthreads):
    (nxdirty, nxfacets), (nydirty, nyfacets) = nx, ny
    if singleprec and epsilon < 1e-6:
        pytest.skip()
    rng = np.random.default_rng(42)
    pixsizex = fov*np.pi/180/nxdirty
    pixsizey = fov*np.pi/180/nydirty*1.1
    f0 = 1e9
    freq = f0 + np.arange(nchan)*(f0/nchan)
    uvw = (rng.random((nrow, 3))-0.5)/(pixsizex*f0/SPEEDOFLIGHT)
    ms = rng.random((nrow, nchan))-0.5 + 1j*(rng.random((nrow, nchan))-0.5)
    wgt = rng.uniform(0.9, 1.1, (nrow, nchan)) if use_wgt else None
    mask = (rng.uniform(0, 1, (nrow, nchan)) > 0.5).astype(np.uint8) \
        if use_mask else None
    nu = nv = 0
    if singleprec:
        ms = ms.astype("c8")
        if wgt is not None:
            wgt = wgt.astype("f4")
    dirty = ng.ms2dirty(uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
                        pixsizey, nu, nv, epsilon, wstacking, nthreads,
                        0, mask).astype("f8")
    ref = explicit_gridder(uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
                           pixsizey, wstacking, mask)
    assert_allclose(ducc0.misc.l2error(dirty, ref), 0, atol=epsilon)

    dirty2 = vis2dirty_with_faceting(nfacets_x=nxfacets, nfacets_y=nyfacets,
                                     uvw=uvw, freq=freq,
                                     vis=ms, wgt=wgt, npix_x=nxdirty,
                                     npix_y=nydirty, pixsize_x=pixsizex,
                                     pixsize_y=pixsizey, epsilon=epsilon,
                                     do_wgridding=wstacking, nthreads=nthreads,
                                     mask=mask).astype("f8")
    assert_allclose(ducc0.misc.l2error(dirty2, ref), 0, atol=epsilon)

    if wstacking or (not have_finufft):
        return
    dirty = with_finufft(uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
                         pixsizey, mask, epsilon)
    assert_allclose(ducc0.misc.l2error(dirty, ref), 0, atol=epsilon)


@pmp('nx', [128])
@pmp('ny', [250])
@pmp("nrow", (27,))
@pmp("nchan", (5,))
@pmp("epsilon", (2e-10,))
@pmp("singleprec", (False,))
@pmp("wstacking", (True,))
@pmp("use_wgt", (True, False))
@pmp("use_mask", (False, True))
@pmp("nthreads", (2,))
@pmp("fov", (10., 120.))
def test_adjointness_wsclean(nx, ny, nrow, nchan, epsilon,
                             singleprec, wstacking, use_wgt, nthreads,
                             use_mask, fov):
    nxdirty, nydirty = nx, ny
    if singleprec and epsilon < 1e-6:
        pytest.skip()
    rng = np.random.default_rng(43)
    pixsizex = fov*np.pi/180/nxdirty
    pixsizey = fov*np.pi/180/nydirty*1.1
    f0 = 1e9
    freq = f0 + np.arange(nchan)*(f0/nchan)
    uvw = (rng.random((nrow, 3))-0.5)/(pixsizey*f0/SPEEDOFLIGHT)
    ms = rng.random((nrow, nchan))-0.5 + 1j*(rng.random((nrow, nchan))-0.5)
    wgt = rng.uniform(0.9, 1.1, (nrow, nchan)) if use_wgt else None
    mask = (rng.uniform(0, 1, (nrow, nchan)) > 0.5).astype(np.uint8) \
        if use_mask else None
    dirty = rng.random((nxdirty, nydirty))-0.5
    nu = nv = 0
    if singleprec:
        ms = ms.astype("c8")
        dirty = dirty.astype("f4")
        if wgt is not None:
            wgt = wgt.astype("f4")

    def check(d2, m2):
        ref = max(vdot(ms, ms).real, vdot(m2, m2).real,
                  vdot(dirty, dirty).real, vdot(d2, d2).real)
        tol = 3e-5*ref if singleprec else 2e-13*ref
        assert_allclose(vdot(ms, m2).real, vdot(d2, dirty), rtol=3*tol)

    dirty2 = ng.vis2dirty(
        uvw=uvw, freq=freq, vis=ms, wgt=wgt, npix_x=nxdirty, npix_y=nydirty,
        pixsize_x=pixsizex, pixsize_y=pixsizey, epsilon=epsilon,
        do_wgridding=wstacking, nthreads=nthreads, mask=mask,
        divide_by_n=False).astype("f8")
    ms2 = ng.dirty2vis(uvw=uvw, freq=freq, dirty=dirty, wgt=wgt,
        pixsize_x=pixsizex, pixsize_y=pixsizey, epsilon=epsilon,
        do_wgridding=wstacking, nthreads=nthreads, mask=mask,
        divide_by_n=False).astype("c16")
    check(dirty2, ms2)


@pmp('nx', [50])
@pmp('ny', [50])
@pmp("nrow", (27,))
@pmp("nchan", (5,))
@pmp("epsilon", (1e-7,))
@pmp("singleprec", (False,))
@pmp("wstacking", (True,))
@pmp("use_wgt", (True,))
@pmp("use_mask", (False,))
@pmp("nthreads", (1,))
@pmp("fov", (10., 80., 90., 120.,))
def test_vis2dirty_wsclean(nx, ny, nrow, nchan, epsilon,
                           singleprec, wstacking, use_wgt, use_mask, fov,
                           nthreads):
    nxdirty, nydirty = nx, ny
    if singleprec and epsilon < 1e-6:
        pytest.skip()
    rng = np.random.default_rng(42)
    pixsizex = fov*np.pi/180/nxdirty
    pixsizey = fov*np.pi/180/nydirty*1.1
    f0 = 1e9
    freq = f0 + np.arange(nchan)*(f0/nchan)
    uvw = (rng.random((nrow, 3))-0.5)/(pixsizex*f0/SPEEDOFLIGHT)
    ms = rng.random((nrow, nchan))-0.5 + 1j*(rng.random((nrow, nchan))-0.5)
    wgt = rng.uniform(0.9, 1.1, (nrow, nchan)) if use_wgt else None
    mask = (rng.uniform(0, 1, (nrow, nchan)) > 0.5).astype(np.uint8) \
        if use_mask else None
    if singleprec:
        ms = ms.astype("c8")
        if wgt is not None:
            wgt = wgt.astype("f4")
    dirty = ng.vis2dirty(
        uvw=uvw, freq=freq, vis=ms, wgt=wgt, npix_x=nxdirty, npix_y=nydirty,
        pixsize_x=pixsizex, pixsize_y=pixsizey, epsilon=epsilon,
        do_wgridding=wstacking, nthreads=nthreads, verbosity=0, mask=mask,
        divide_by_n=False).astype("f8")
    ref = explicit_gridder(uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
                           pixsizey, wstacking, mask, divide_by_n=False)
    dirty[0,0]=ref[0,0]
    assert_allclose(ducc0.misc.l2error(dirty, ref), 0, atol=epsilon)


@pmp('nxdirty', [2, 16, 64])
@pmp('nydirty', [2, 64])
@pmp("nrow", (1, 100))
@pmp("nchan", (1, 7))
@pmp("epsilon", list(10.**np.linspace(-2., -12., 20)))
@pmp("singleprec", (False,))
@pmp("wstacking", (True,))
@pmp("use_wgt", (True,))
@pmp("nthreads", (1, 10))
@pmp("fov", (10.,))
def test_ms2dirty_against_wdft3(nxdirty, nydirty, nrow, nchan, epsilon,
                                singleprec, wstacking, use_wgt, fov, nthreads):
    if singleprec and epsilon < 5e-5:
        return
    rng = np.random.default_rng(42)
    pixsizex = fov*np.pi/180/nxdirty
    pixsizey = fov*np.pi/180/nydirty*1.1
    speedoflight, f0 = 299792458., 1e9
    freq = f0 + np.arange(nchan)*(f0/nchan)
    uvw = (rng.random((nrow, 3))-0.5)/(pixsizex*f0/speedoflight)
    ms = rng.random((nrow, nchan))-0.5 + 1j*(rng.random((nrow, nchan))-0.5)
    wgt = rng.random((nrow, 1)) if use_wgt else None
    wgt = np.broadcast_to(wgt, (nrow, nchan)) if use_wgt else None
    if singleprec:
        ms = ms.astype("c8")
        if wgt is not None:
            wgt = wgt.astype("f4")
    try:
        dirty = ng.ms2dirty(
            uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
            pixsizey, 0, 0, epsilon, wstacking, nthreads, 0).astype("f8")
    except:
        # no matching kernel was found
        pytest.skip()
    ref = explicit_gridder(uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
                           pixsizey, wstacking, None)
    assert_allclose(ducc0.misc.l2error(dirty, ref), 0, atol=2*epsilon)
    x1 = explicit_degridder(uvw, freq, ref, wgt, pixsizex,
                           pixsizey, wstacking, None)
    x2 = ng.dirty2ms(uvw, freq, ref, wgt, pixsizex, pixsizey, 0, 0,
                      epsilon, wstacking, nthreads, 0).astype("c16")
    assert_allclose(ducc0.misc.l2error(x1,x2), 0, atol=epsilon)


@pmp('nx', [(2, 2), (30, 3), (128, 2)])
@pmp('ny', [(2, 2), (128, 2), (250, 5)])
@pmp("nrow", (1, 2, 27))
@pmp("nchan", (1, 5))
@pmp("epsilon", (1e-1, 3e-5, 2e-13))
@pmp("singleprec", (True, False))
@pmp("wstacking", (True, False))
@pmp("use_wgt", (True, False))
@pmp("use_mask", (False, True))
@pmp("nthreads", (1, 2))
def test_adjointness_ms2dirty_complex(nx, ny, nrow, nchan, epsilon,
                              singleprec, wstacking, use_wgt, nthreads,
                              use_mask):
    (nxdirty, nxfacets), (nydirty, nyfacets) = nx, ny
    if singleprec and epsilon < 1e-6:
        pytest.skip()
    rng = np.random.default_rng(42)
    pixsizex = np.pi/180/60/nxdirty*0.2398
    pixsizey = np.pi/180/60/nxdirty
    f0 = 1e9
    freq = f0 + np.arange(nchan)*(f0/nchan)
    uvw = (rng.random((nrow, 3))-0.5)/(pixsizey*f0/SPEEDOFLIGHT)
    ms = rng.random((nrow, nchan))-0.5 + 1j*(rng.random((nrow, nchan))-0.5)
    wgt = rng.uniform(0.9, 1.1, (nrow, nchan)) if use_wgt else None
    mask = (rng.uniform(0, 1, (nrow, nchan)) > 0.5).astype(np.uint8) \
        if use_mask else None
    dirty = rng.random((nxdirty, nydirty))-0.5
    dirty = dirty +  1j*(rng.random((nxdirty, nydirty))-0.5)
    nu = nv = 0
    if singleprec:
        ms = ms.astype("c8")
        dirty = dirty.astype("c8")
        if wgt is not None:
            wgt = wgt.astype("f4")

    def check(d2, m2):
        ref = max(vdot(ms, ms).real, vdot(m2, m2).real,
                  vdot(dirty, dirty).real, vdot(d2, d2).real)
        tol = 3e-5*ref if singleprec else 2e-13*ref
        assert_allclose(vdot(ms, m2), vdot(d2, dirty), rtol=tol)

    dirty2 = ng.ms2dirty(uvw, freq, ms, wgt, nxdirty, nydirty, pixsizex,
                         pixsizey, nu, nv, epsilon, wstacking, nthreads, 0,
                         mask).astype("f8") \
            +1j * ng.ms2dirty(uvw, freq, -1j*ms, wgt, nxdirty, nydirty, pixsizex,
                         pixsizey, nu, nv, epsilon, wstacking, nthreads, 0,
                         mask).astype("f8")
    ms2 = ng.dirty2ms(uvw, freq, dirty.real, wgt, pixsizex, pixsizey, nu, nv,
                       epsilon, wstacking, nthreads, 0, mask).astype("c16") \
          +1j*ng.dirty2ms(uvw, freq, dirty.imag, wgt, pixsizex, pixsizey, nu, nv,
                       epsilon, wstacking, nthreads, 0, mask).astype("c16")
    check(dirty2, ms2)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020 Max-Planck-Society


import ducc0
import ducc0.pointingprovider as pp
# import pyfftw
import numpy as np
import pytest
from numpy.testing import assert_, assert_allclose


pmp = pytest.mark.parametrize


@pmp("size", (2, 10, 37, 1000))
@pmp("t0", (-45.3, 0, 10))
@pmp("freq", (1, 1.3e-7, 3e10))
@pmp("nthreads", (1, 2))
def testp1(size, t0, freq, nthreads):
    rng = np.random.default_rng(42)
    quat = rng.uniform(-.5, .5, (size, 4))
    prov = pp.PointingProvider(t0, freq, quat, nthreads)
    rquat = np.array([0., 0., 0., 1.])  # a non-rotating quaternion
    quat2 = prov.get_rotated_quaternions(t0, freq, rquat, size)
    quat3 = prov.get_rotated_quaternions(t0, freq, rquat, size, rot_left=False)
    nquat = quat/np.sqrt(np.sum(quat**2, axis=1, keepdims=True))
    # adjust signs
    nquat = nquat * np.sign(nquat[:, 0]).reshape(-1, 1)
    quat2 = quat2 * np.sign(quat2[:, 0]).reshape(-1, 1)
    quat3 = quat3 * np.sign(quat3[:, 0]).reshape(-1, 1)
    assert_allclose(ducc0.misc.l2error(quat2, nquat), 0, atol=1e-13)
    assert_allclose(ducc0.misc.l2error(quat3, nquat), 0, atol=1e-13)


def testp2():
    rng = np.random.default_rng(42)
    t01, f1, size1 = 0., 1., 200
    quat1 = rng.uniform(-.5, .5, (size1, 4))
    prov = pp.PointingProvider(t01, f1, quat1)
    rquat = rng.uniform(-.5, .5, (4,))
    t02, f2, size2 = 3.7, 10.2, 300
    quat2 = prov.get_rotated_quaternions(t02, f2, rquat, size2)
    quat3 = np.empty((size2, 4), dtype=np.float64)
    quat3 = prov.get_rotated_quaternions(t02, f2, rquat, out=quat3)
    quat4 = np.empty((size2, 4), dtype=np.float64)
    quat4 = prov.get_rotated_quaternions(t02, f2, rquat, rot_left=False,
                                         out=quat4)
    assert_((quat2 == quat3).all(), "problem")
    quat2 *= np.sign(quat2[:, 0]).reshape((-1, 1))
    quat4 *= np.sign(quat4[:, 0]).reshape((-1, 1))

    try:
        from scipy.spatial.transform import Rotation as R
        from scipy.spatial.transform import Slerp
    except:
        pytest.skip()
    times1 = t01 + 1./f1*np.arange(size1)
    r1 = R.from_quat(quat1)
    rrquat = R.from_quat(rquat)
    slerp = Slerp(times1, r1)
    times2 = t02 + 1./f2*np.arange(size2)
    r2 = rrquat*slerp(times2)
    squat2 = r2.as_quat()
    squat2 *= np.sign(squat2[:, 0]).reshape((-1, 1))
    assert_allclose(ducc0.misc.l2error(quat2, squat2), 0, atol=1e-13)
    r3 = slerp(times2)*rrquat
    squat3 = r3.as_quat()
    squat3 *= np.sign(squat3[:, 0]).reshape((-1, 1))
    assert_allclose(ducc0.misc.l2error(quat4, squat3), 0, atol=1e-13)


def testp3():  # test periodicity
    rng = np.random.default_rng(42)
    t01, f1, size1 = 0., 1., 20
    quat1 = rng.uniform(-.5, .5, (size1, 4))
    prov = pp.PointingProvider(t01, f1, quat1)
    rquat = rng.uniform(-.5, .5, (4,))
    t02, f2, size2 = 3.7, 10.2, 30000
    quat2 = prov.get_rotated_quaternions(t02, f2, rquat, size2)
    t03 = t02 + 2*size1/f1
    quat3 = prov.get_rotated_quaternions(t03, f2, rquat, size2)
    assert_allclose(ducc0.misc.l2error(quat2, quat3), 0., atol=1e-13)
    t04 = t02 - 5*size1/f1
    quat4 = prov.get_rotated_quaternions(t04, f2, rquat, size2)
    assert_allclose(ducc0.misc.l2error(quat2, quat4), 0., atol=1e-13)


def test_quat_euler():
    nsamp = 10000

    euler = np.empty((nsamp,3))

    # theta
    euler[:,0] = np.random.uniform(low=-3*np.pi, high=5*np.pi, size=nsamp)
    # phi
    euler[:,1] = np.random.uniform(low=-5*np.pi, high=8*np.pi, size=nsamp)
    # psi
    euler[:,2] = np.random.uniform(low=-5*np.pi, high=8*np.pi, size=nsamp)

    euler2 = ducc0.misc.quat2ptg(ducc0.misc.ptg2quat(euler))

    # Normalize angle ranges
    euler = (euler + 8*np.pi)%(2*np.pi)
    mask = euler[:,0]>np.pi
    euler[:,0] = np.where(mask,2*np.pi-euler[:,0],euler[:,0])
    euler[:,1] = np.where(mask,(euler[:,1]+np.pi)%(2*np.pi),euler[:,1])
    euler[:,2] = np.where(mask,(euler[:,2]+np.pi)%(2*np.pi),euler[:,2])

    euler2[:,1:3] = (euler2[:,1:3] + 8*np.pi)%(2*np.pi)

    assert_allclose(ducc0.misc.l2error(euler, euler2), 0, atol=1e-13)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2025 Max-Planck-Society
# Copyright(C) 2025 Philipp Arras


import ducc0
import numpy as np
import pytest
from numpy.testing import assert_allclose

pmp = pytest.mark.parametrize


@pmp("shape", ([43], [654, 23], [32, 3, 11]))
@pmp("dtype_cov", (np.float32, np.float64))
@pmp("cplx", (False, True))
@pmp("broadcast", (False, True))
@pmp("nthreads", (1, 2))
def test_gaussenergy(shape, dtype_cov, cplx, broadcast, nthreads):
    rng = np.random.default_rng(42)

    a = rng.uniform(-.5, .5, shape).astype(dtype_cov)
    b = rng.uniform(-.5, .5, shape).astype(dtype_cov)
    c = rng.uniform(-.5, .5, shape).astype(dtype_cov)
    if cplx:
        a = a + 1j*rng.uniform(-.5, .5, shape).astype(dtype_cov)
        b = b + 1j*rng.uniform(-.5, .5, shape).astype(dtype_cov)
    if broadcast:
        a = np.broadcast_to(a[2:3], b.shape)
    res = ducc0.misc.experimental.LogUnnormalizedGaussProbability(a, b, c, nthreads)
    ref = 0.5*ducc0.misc.vdot((a-b)*c, a-b).real
    rtol = 1e-5 if dtype_cov == np.float32 else 1e-12
    assert_allclose(res, ref, rtol=rtol)

    res, deriv = ducc0.misc.experimental.LogUnnormalizedGaussProbabilityWithDeriv(a, b, c, nthreads=nthreads)
    assert_allclose(res, ref, rtol=rtol)
    assert_allclose(deriv, (a-b)*c, rtol=rtol)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020-2021 Max-Planck-Society


import numpy as np
import pytest
from numpy.testing import assert_allclose
import ducc0

pmp = pytest.mark.parametrize


def _assert_close(a, b, epsilon):
    if isinstance(a, np.ndarray):
        err = ducc0.misc.l2error(a, b)
    else:
        err = np.abs(a-b)/np.maximum(np.abs(a),np.abs(b))
    if (err >= epsilon):
        print("Error: {} > {}".format(err, epsilon))
    assert_allclose(err, 0, atol=epsilon)


def nalm(lmax, mmax):
    return ((mmax+1)*(mmax+2))//2 + (mmax+1)*(lmax-mmax)


def random_alm(rng, lmax, mmax, ncomp):
    res = rng.uniform(-1., 1., (ncomp, nalm(lmax, mmax))) \
     + 1j*rng.uniform(-1., 1., (ncomp, nalm(lmax, mmax)))
    # make a_lm with m==0 real-valued
    res[:, 0:lmax+1].imag = 0.
    return res


def convolve(alm1, alm2, lmax, nthreads=1):
    ntheta, nphi = lmax+1, ducc0.fft.good_size(2*lmax+1, True)
    tmap = ducc0.sht.synthesis_2d(
        alm=alm1.reshape((1,-1)), ntheta=ntheta, nphi=nphi, lmax=lmax,
        geometry="GL", spin=0, nthreads=nthreads)
    tmap *= ducc0.sht.synthesis_2d(
        alm=alm2.reshape((1,-1)), ntheta=ntheta, nphi=nphi, lmax=lmax,
        geometry="GL", spin=0, nthreads=nthreads)
    res = ducc0.sht.analysis_2d(
        map=tmap, lmax=0, spin=0, geometry="GL", nthreads=nthreads)
    return np.sqrt(4*np.pi)*res[0,0]


def compress_alm(alm, lmax):
    res = np.empty(2*len(alm)-lmax-1, dtype=np.float64)
    res[0:lmax+1] = alm[0:lmax+1].real
    res[lmax+1::2] = np.sqrt(2)*alm[lmax+1:].real
    res[lmax+2::2] = np.sqrt(2)*alm[lmax+1:].imag
    return res


def myalmdot(a1, a2, lmax):
    return ducc0.misc.vdot(compress_alm(a1, lmax), compress_alm(a2, lmax))


@pmp("lkmax", [(13, 13), (2, 1), (30, 15), (35, 2)])
def test_against_convolution(lkmax):
    lmax, kmax = lkmax
    rng = np.random.default_rng(42)
    slm = random_alm(rng, lmax, lmax, 1)[0, :]
    blm = random_alm(rng, lmax, kmax, 1)[0, :]

    conv = ducc0.totalconvolve.ConvolverPlan(lmax, kmax,
                                             epsilon=1.5e-13, nthreads=2)
    nptg = 50
    ptg = np.zeros((nptg, 3))
    ptg[:, 0] = rng.uniform(0, np.pi, nptg)
    ptg[:, 1] = rng.uniform(0, 2*np.pi, nptg)
    ptg[:, 2] = rng.uniform(0, 2*np.pi, nptg)

    cube = np.empty((conv.Npsi(), conv.Ntheta(), conv.Nphi()))
    conv.getPlane(slm, blm, 0, cube[0:1])
    for mbeam in range(1, kmax+1):
        conv.getPlane(slm, blm, mbeam, cube[2*mbeam-1:2*mbeam+1])
    conv.prepPsi(cube)
    res1 = np.empty(ptg.shape[0])
    conv.interpol(cube, 0, 0, ptg[:, 0], ptg[:, 1], ptg[:, 2], res1)

    blm2 = np.zeros((nalm(lmax, lmax),))+0j
    blm2[0:blm.shape[0]] = blm
    res2 = np.zeros((nptg,))
    for i in range(nptg):
        rbeam = ducc0.sht.rotate_alm(blm2, lmax, ptg[i, 2], ptg[i, 0], ptg[i, 1])
        res2[i] = convolve(slm, rbeam, lmax).real
    _assert_close(res1, res2, 1.5e-13)


@pmp("lkmax", [(13, 13), (2, 1), (30, 15), (35, 2)])
@pmp("ncomp", [1, 3])
@pmp("separate", [True, False])
def test_against_convolution_2(lkmax, ncomp, separate):
    lmax, kmax = lkmax
    rng = np.random.default_rng(42)
    slm = random_alm(rng, lmax, lmax, ncomp)
    blm = random_alm(rng, lmax, kmax, ncomp)

    inter = ducc0.totalconvolve.Interpolator(slm, blm, separate, lmax, kmax,
                                             epsilon=1.5e-13, nthreads=2)
    nptg = 50
    ptg = np.zeros((nptg, 3))
    ptg[:, 0] = rng.uniform(0, np.pi, nptg)
    ptg[:, 1] = rng.uniform(0, 2*np.pi, nptg)
    ptg[:, 2] = rng.uniform(-np.pi, np.pi, nptg)

    res1 = inter.interpol(ptg)

    blm2 = np.zeros((ncomp, nalm(lmax, lmax)))+0j
    blm2[:, 0:blm.shape[1]] = blm
    res2 = np.zeros((ncomp, nptg))
    for c in range(ncomp):
        for i in range(nptg):
            rbeam = ducc0.sht.rotate_alm(blm2[c, :], lmax,
                                         ptg[i, 2], ptg[i, 0], ptg[i, 1])
            res2[c, i] = convolve(slm[c, :], rbeam, lmax).real
    if separate:
        _assert_close(res1, res2, 1.5e-13)
    else:
        _assert_close(res1[0, :], np.sum(res2, axis=0), 1.5e-13)


@pmp("lkmax", [(13, 13), (2, 1), (30, 15), (35, 2)])
@pmp("ncomp", [1, 3])
@pmp("separate", [True, False])
def test_against_convolution_2f(lkmax, ncomp, separate):
    lmax, kmax = lkmax
    rng = np.random.default_rng(42)
    slm = random_alm(rng, lmax, lmax, ncomp).astype("c8")
    blm = random_alm(rng, lmax, kmax, ncomp).astype("c8")

    inter = ducc0.totalconvolve.Interpolator_f(slm, blm, separate, lmax, kmax,
                                               epsilon=4e-6, nthreads=2)
    nptg = 50
    ptg = np.zeros((nptg, 3))
    ptg[:, 0] = rng.uniform(0, np.pi, nptg)
    ptg[:, 1] = rng.uniform(0, 2*np.pi, nptg)
    ptg[:, 2] = rng.uniform(-np.pi, np.pi, nptg)
    ptg = ptg.astype("f4")

    res1 = inter.interpol(ptg)

    blm2 = np.zeros((ncomp, nalm(lmax, lmax)))+0j
    blm2[:, 0:blm.shape[1]] = blm
    res2 = np.zeros((ncomp, nptg))
    for c in range(ncomp):
        for i in range(nptg):
            rbeam = ducc0.sht.rotate_alm(blm2[c, :], lmax,
                                         ptg[i, 2], ptg[i, 0], ptg[i, 1])
            res2[c, i] = convolve(slm[c, :], rbeam, lmax).real
    if separate:
        _assert_close(res1, res2, 4e-6)
    else:
        _assert_close(res1[0, :], np.sum(res2, axis=0), 4e-6)


@pmp("lkmax", [(13, 13), (20, 0), (2, 1), (30, 15), (35, 2)])
def test_adjointness(lkmax):
    lmax, kmax = lkmax
    rng = np.random.default_rng(42)
    slm = random_alm(rng, lmax, lmax, 1)[0, :]
    blm = random_alm(rng, lmax, kmax, 1)[0, :]
    nptg = 50
    ptg = rng.uniform(0., 1., nptg*3).reshape(nptg, 3)
    ptg[:, 0] *= np.pi
    ptg[:, 1] *= 2*np.pi
    ptg[:, 2] *= 2*np.pi
    conv = ducc0.totalconvolve.ConvolverPlan(lmax, kmax,
                                             epsilon=4e-6, nthreads=2)

    cube = np.empty((conv.Npsi(), conv.Ntheta(), conv.Nphi()))
    conv.getPlane(slm, blm, 0, cube[0:1])
    for mbeam in range(1, kmax+1):
        conv.getPlane(slm, blm, mbeam, cube[2*mbeam-1:2*mbeam+1])
    conv.prepPsi(cube)
    inter1 = np.empty(ptg.shape[0])
    conv.interpol(cube, 0, 0, ptg[:, 0], ptg[:, 1], ptg[:, 2], inter1)

    fake = rng.uniform(-0.5, 0.5, (ptg.shape[0],))
    cube2 = cube*0.
    conv.deinterpol(cube2, 0, 0, ptg[:, 0], ptg[:, 1], ptg[:, 2], fake)
    bla = slm*0.
    conv.deprepPsi(cube2)
    conv.updateSlm(bla, blm, 0, cube2[0:1])
    for mbeam in range(1, kmax+1):
        conv.updateSlm(bla, blm, mbeam, cube2[2*mbeam-1:2*mbeam+1])

    v1 = myalmdot(slm, bla, lmax)
    v2 = ducc0.misc.vdot(fake, inter1)
    _assert_close(v1, v2, 1e-13)


@pmp("lkmax", [(13, 13), (2, 1), (30, 15), (35, 2)])
@pmp("ncomp", [1, 3])
@pmp("separate", [True, False])
@pmp("single", [True, False])
def test_adjointness2(lkmax, ncomp, separate, single):
    lmax, kmax = lkmax
    rng = np.random.default_rng(42)
    slm = random_alm(rng, lmax, lmax, ncomp)
    blm = random_alm(rng, lmax, kmax, ncomp)
    nptg = 50
    ptg = rng.uniform(0., 1., nptg*3).reshape(nptg, 3)
    ptg[:, 0] *= np.pi
    ptg[:, 1] *= 2*np.pi
    ptg[:, 2] *= 2*np.pi
    if single:
        slm = slm.astype("c8")
        blm = blm.astype("c8")
        ptg = ptg.astype("f4")
        foo = ducc0.totalconvolve.Interpolator_f(slm, blm, separate, lmax, kmax,
                                                 epsilon=1e-5, nthreads=2)
    else:
        foo = ducc0.totalconvolve.Interpolator(slm, blm, separate, lmax, kmax,
                                               epsilon=1e-6, nthreads=2)
    inter1 = foo.interpol(ptg).astype("f8")
    ncomp2 = inter1.shape[0]
    fake = rng.uniform(-0.5, 0.5, (ncomp2, ptg.shape[0]))
    if single:
        fake = fake.astype("f4")
        foo2 = ducc0.totalconvolve.Interpolator_f(lmax, kmax, ncomp2, epsilon=1e-5,
                                                  nthreads=2)
    else:
        foo2 = ducc0.totalconvolve.Interpolator(lmax, kmax, ncomp2, epsilon=1e-6,
                                                nthreads=2)
    foo2.deinterpol(ptg.reshape((-1, 3)), fake)
    bla = foo2.getSlm(blm).astype("c16")
    v1 = np.sum([myalmdot(slm[c, :], bla[c, :], lmax)
                 for c in range(ncomp)])
    v2 = np.sum([ducc0.misc.vdot(fake[c, :], inter1[c, :]) for c in range(ncomp2)])
    epsilon = 1e-4 if single else 1e-12
    _assert_close(v1, v2, epsilon)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2020-2023 Max-Planck-Society


import ducc0
import numpy as np
import pytest
from numpy.testing import assert_allclose

pmp = pytest.mark.parametrize


def nalm(lmax, mmax):
    return ((mmax+1)*(mmax+2))//2 + (mmax+1)*(lmax-mmax)


def random_alm(lmax, mmax, spin, ncomp, rng):
    res = rng.uniform(-1., 1., (ncomp, nalm(lmax, mmax))) \
     + 1j*rng.uniform(-1., 1., (ncomp, nalm(lmax, mmax)))
    # make a_lm with m==0 real-valued
    res[:, 0:lmax+1].imag = 0.
    ofs=0
    for s in range(spin):
        res[:, ofs:ofs+spin-s] = 0.
        ofs += lmax+1-s
    return res


def compress_alm(alm, lmax):
    res = np.empty(2*len(alm)-lmax-1, dtype=np.float64)
    res[0:lmax+1] = alm[0:lmax+1].real
    res[lmax+1::2] = np.sqrt(2)*alm[lmax+1:].real
    res[lmax+2::2] = np.sqrt(2)*alm[lmax+1:].imag
    return res


def myalmdot(a1, a2, lmax):
    return ducc0.misc.vdot(compress_alm(a1, lmax), compress_alm((a2), lmax))


@pmp('geometry', ("CC", "F1", "MW", "MWflip", "GL", "DH", "F2"))
@pmp('spin', (0, 1, 2))
@pmp('nthreads', (1, 4))
@pmp('lmmax', ((2, 0), (2, 2), (5, 5), (11, 10), (11, 11), (32, 32), (200, 200)))
@pmp('phi0', (0, 0.03))
def test_2d_roundtrip(lmmax, geometry, spin, nthreads, phi0):
    rng = np.random.default_rng(np.random.SeedSequence(42))
    ncomp = 1 if spin == 0 else 2
    lmax, mmax = lmmax

    nrings = lmax+1
    if geometry=="CC":
        nrings = lmax+2
    elif geometry=="DH":
        nrings = 2*lmax+2
    elif geometry=="F2":
        nrings = 2*lmax+1
    nphi=2*mmax+2

    alm = random_alm(lmax, mmax, spin, ncomp, rng)
    map = ducc0.sht.synthesis_2d(alm=alm, lmax=lmax, mmax=mmax, spin=spin, ntheta=nrings, nphi=nphi, nthreads=nthreads, geometry=geometry, phi0=phi0)
    alm2 = ducc0.sht.analysis_2d(map=map, lmax=lmax, mmax=mmax, spin=spin, geometry=geometry, phi0=phi0, nthreads=nthreads)
    assert_allclose(ducc0.misc.l2error(alm2,alm), 0, atol=1e-12)

    map = ducc0.sht.adjoint_analysis_2d(alm=alm, lmax=lmax, mmax=mmax, spin=spin, ntheta=nrings, nphi=nphi, nthreads=nthreads, geometry=geometry, phi0=phi0)
    alm2 = ducc0.sht.adjoint_synthesis_2d(map=map, lmax=lmax, mmax=mmax, spin=spin, geometry=geometry, phi0=phi0, nthreads=nthreads)
    assert_allclose(ducc0.misc.l2error(alm2,alm), 0, atol=1e-12)

    if phi0 != 0: return

    job = ducc0.sht.sharpjob_d()
    job.set_triangular_alm_info(lmax, mmax)
    if geometry=="CC":
        job.set_cc_geometry(nrings, nphi)
    elif geometry=="DH":
        job.set_dh_geometry(nrings, nphi)
    elif geometry=="F1":
        job.set_fejer1_geometry(nrings, nphi)
    elif geometry=="F2":
        job.set_fejer2_geometry(nrings, nphi)
    elif geometry=="MW":
        job.set_mw_geometry(nrings, nphi)
    elif geometry=="GL":
        job.set_gauss_geometry(nrings, nphi)
    else:
        return

    if spin == 0:
        alm2 = job.map2alm(job.alm2map(alm[0])).reshape((1,-1))
    else:
        alm2 = job.map2alm_spin(job.alm2map_spin(alm, spin), spin)
    assert_allclose(alm, alm2)


@pmp('geometry', ("CC", "F1", "MW", "MWflip", "GL", "DH", "F2"))
@pmp('spin', (0, 1, 2))
@pmp('nthreads', (1, 4))
@pmp('lmmax', ((2, 0), (2, 2), (5, 5), (11, 10), (11, 11), (32, 32), (200, 200)))
def test_2d_adjoint(lmmax, geometry, spin, nthreads):
    rng = np.random.default_rng(48)

    lmax, mmax = lmmax
    ncomp = 1 if spin == 0 else 2

    nrings = lmax+1
    if geometry=="CC":
        nrings = lmax+2
    elif geometry=="DH":
        nrings = 2*lmax+2
    elif geometry=="F2":
        nrings = 2*lmax+1
    nphi=2*mmax+2

    alm0 = random_alm(lmax, mmax, spin, ncomp, rng)
    map0 = rng.uniform(0., 1., (alm0.shape[0], nrings, nphi))

    # test adjointness between synthesis and adjoint_synthesis
    map1 = ducc0.sht.synthesis_2d(alm=alm0, lmax=lmax, mmax=mmax, spin=spin, ntheta=nrings, nphi=nphi, nthreads=nthreads, geometry=geometry)
    v2 = np.sum([ducc0.misc.vdot(map0[i], map1[i]) for i in range(ncomp)])
    del map1
    alm1 = ducc0.sht.adjoint_synthesis_2d(lmax=lmax, mmax=mmax, spin=spin, map=map0, nthreads=nthreads, geometry=geometry)
    v1 = np.sum([myalmdot(alm0[i], alm1[i], lmax) for i in range(ncomp)])
    assert_allclose(v1, v2, rtol=1e-10)

    if spin > 0:
        # test adjointness between synthesis and adjoint_synthesis (gradient only)
        map1 = ducc0.sht.synthesis_2d(alm=alm0[:1], lmax=lmax, mmax=mmax, spin=spin, ntheta=nrings, nphi=nphi, nthreads=nthreads, geometry=geometry, mode="GRAD_ONLY")
        v2 = np.sum([ducc0.misc.vdot(map0[i], map1[i]) for i in range(ncomp)])
        # compare grad-only alm2map with full alm2map where curl is set to zero
        almx = alm0.copy()
        almx[1] = 0
        map1b = ducc0.sht.synthesis_2d(alm=almx, lmax=lmax, mmax=mmax, spin=spin, ntheta=nrings, nphi=nphi, nthreads=nthreads, geometry=geometry)
        assert_allclose(ducc0.misc.l2error(map1, map1b), 0, atol=1e-10)
        del almx, map1
        alm1 = ducc0.sht.adjoint_synthesis_2d(lmax=lmax, mmax=mmax, spin=spin, map=map0, nthreads=nthreads, geometry=geometry, mode="GRAD_ONLY")
        v1 = np.sum([myalmdot(alm0[i], alm1[i], lmax) for i in range(1)])
        assert_allclose(np.abs((v1-v2)/v1), 0, atol=1e-10)

    # test adjointness between analysis and adjoint_analysis
    map1 = ducc0.sht.adjoint_analysis_2d(alm=alm0, lmax=lmax, mmax=mmax, spin=spin, ntheta=nrings, nphi=nphi, nthreads=nthreads, geometry=geometry)
    v2 = np.sum([ducc0.misc.vdot(map0[i], map1[i]) for i in range(ncomp)])
    del map1
    alm1 = ducc0.sht.analysis_2d(lmax=lmax, mmax=mmax, spin=spin, map=map0, nthreads=nthreads, geometry=geometry)
    v1 = np.sum([myalmdot(alm0[i], alm1[i], lmax) for i in range(ncomp)])
    assert_allclose(v1, v2, rtol=1e-10)


@pmp('spin', (0, 1, 2))
@pmp('nthreads', (1, 4))
@pmp('lmax', (2, 5, 32, 256))
@pmp('mmaxhalf', (False, True))
@pmp('nside', (2, 5, 128))
@pmp('theta_interpol', (False, True))
def test_healpix_adjoint(lmax, nside, spin, mmaxhalf, theta_interpol, nthreads):
    rng = np.random.default_rng(48)

    mmax = lmax//2 if mmaxhalf else lmax
    ncomp = 1 if spin == 0 else 2

    alm0 = random_alm(lmax, mmax, spin, ncomp, rng)
    map0 = rng.uniform(0., 1., (alm0.shape[0], 12*nside**2))

    base = ducc0.healpix.Healpix_Base(nside, "RING")
    geom = base.sht_info()

    # test adjointness between synthesis and adjoint_synthesis
    map1 = ducc0.sht.synthesis(alm=alm0, lmax=lmax, spin=spin, nthreads=nthreads, mmax=mmax, theta_interpol=theta_interpol, **geom)
    v2 = np.sum([ducc0.misc.vdot(map0[i], map1[i]) for i in range(ncomp)])
    del map1
    alm1 = ducc0.sht.adjoint_synthesis(lmax=lmax, spin=spin, map=map0, nthreads=nthreads, mmax=mmax, theta_interpol=theta_interpol, **geom)
    v1 = np.sum([myalmdot(alm0[i], alm1[i], lmax) for i in range(ncomp)])
    assert_allclose(v1, v2, rtol=1e-10)

    if spin > 0:
        map1 = ducc0.sht.synthesis(alm=alm0[:1], lmax=lmax, spin=spin, nthreads=nthreads, mmax=mmax, mode="GRAD_ONLY", theta_interpol=theta_interpol, **geom)
        v2 = np.sum([ducc0.misc.vdot(map0[i], map1[i]) for i in range(ncomp)])
        del map1
        alm1 = ducc0.sht.adjoint_synthesis(lmax=lmax, spin=spin, map=map0, nthreads=nthreads, mmax=mmax, mode="GRAD_ONLY", theta_interpol=theta_interpol, **geom)
        v1 = np.sum([myalmdot(alm0[i], alm1[i], lmax) for i in range(1)])
        assert_allclose(np.abs((v1-v2)/v1), 0, atol=1e-10)


@pmp("lmax", tuple(range(0,70,3)))
@pmp("nthreads", (0,1,2))
def test_rotation(lmax, nthreads):
    rng = np.random.default_rng(42)
    phi, theta, psi = rng.uniform(-2*np.pi, 2*np.pi, (3,))

    alm = random_alm(lmax, lmax, 0, 1, rng)[0,:]
    alm2 = ducc0.sht.rotate_alm(alm, lmax, phi, theta, psi, nthreads)
    alm2 = ducc0.sht.rotate_alm(alm2, lmax, -psi, -theta, -phi, nthreads)
    assert_allclose(ducc0.misc.l2error(alm,alm2), 0, atol=1e-12)
    alm3 = alm.copy()
    alm3 = ducc0.sht.rotate_alm(alm3, lmax, phi, theta, psi, nthreads, out=alm3)
    alm3 = ducc0.sht.rotate_alm(alm3, lmax, -psi, -theta, -phi, nthreads, out=alm3)
    assert_allclose(ducc0.misc.l2error(alm,alm3), 0, atol=1e-12)
    alm = alm.astype(np.complex64)
    alm2 = ducc0.sht.rotate_alm(alm, lmax, phi, theta, psi, nthreads)
    alm2 = ducc0.sht.rotate_alm(alm2, lmax, -psi, -theta, -phi, nthreads)
    assert_allclose(ducc0.misc.l2error(alm,alm2), 0, atol=1e-6)
    alm3 = alm.copy()
    alm3 = ducc0.sht.rotate_alm(alm3, lmax, phi, theta, psi, nthreads, out=alm3)
    alm3 = ducc0.sht.rotate_alm(alm3, lmax, -psi, -theta, -phi, nthreads, out=alm3)
    assert_allclose(ducc0.misc.l2error(alm,alm3), 0, atol=1e-6)


@pmp('spin', (0, 2))
@pmp('nthreads', (1, 4))
@pmp('nside', (32, 64))
@pmp('ntrans', (1, 8))
def test_healpix_inverse(nside, spin, ntrans, nthreads):
    rng = np.random.default_rng(48)

    lmax = int(2.5*nside)
    mmax = lmax
    ncomp = 1 if spin == 0 else 2

    alm0 = random_alm(lmax, mmax, spin, ntrans*ncomp, rng)
    alm0 = alm0.reshape((ntrans, ncomp, -1))

    base = ducc0.healpix.Healpix_Base(nside, "RING")
    geom = base.sht_info()

    map = ducc0.sht.synthesis(alm=alm0, lmax=lmax, spin=spin, nthreads=nthreads, **geom)
    alm1 = ducc0.sht.pseudo_analysis(map=map, lmax=lmax, spin=spin, nthreads=nthreads, epsilon=1e-8, maxiter=20, **geom)[0]

    assert_allclose(ducc0.misc.l2error(alm0,alm1), 0, atol=1e-6)


@pmp('spin', (0, 1, 2))
@pmp('nthreads', (1, 4))
@pmp('npix', (33, 200))
@pmp('lmmax', ((10, 8), (100,100)))
def test_adjointness_general(lmmax, npix, spin, nthreads):
    rng = np.random.default_rng(48)

    lmax, mmax = lmmax
    epsilon = 1e-5
    ncomp = 1 if spin == 0 else 2
    slm1 = random_alm(lmax, mmax, spin, ncomp, rng)
    loc = rng.uniform(0., 1., (npix,2))
    loc[:, 0] *= np.pi
    loc[:, 1] *= 2*np.pi
    points2 = rng.uniform(-0.5, 0.5, (loc.shape[0],ncomp)).T
    points1 = ducc0.sht.synthesis_general(lmax=lmax, mmax=mmax, alm=slm1, loc=loc, spin=spin, epsilon=epsilon, nthreads=nthreads)
    slm2 = ducc0.sht.adjoint_synthesis_general(lmax=lmax, mmax=mmax, map=points2, loc=loc, spin=spin, epsilon=epsilon, nthreads=nthreads)
    v1 = np.sum([myalmdot(slm1[c, :], slm2[c, :], lmax)
                for c in range(ncomp)])
    v2 = ducc0.misc.vdot(points2.real, points1.real) + ducc0.misc.vdot(points2.imag, points1.imag)
    assert_allclose(v1, v2, rtol=1e-9)

    if spin > 0:
        points1 = ducc0.sht.synthesis_general(lmax=lmax, mmax=mmax, alm=slm1[:1], loc=loc, spin=spin, epsilon=epsilon, nthreads=nthreads, mode="GRAD_ONLY")
        slm2 = ducc0.sht.adjoint_synthesis_general(lmax=lmax, mmax=mmax, map=points2, loc=loc, spin=spin, epsilon=epsilon, nthreads=nthreads, mode="GRAD_ONLY")
        v1 = np.sum([myalmdot(slm1[c, :], slm2[c, :], lmax)
                    for c in range(1)])
        v2 = ducc0.misc.vdot(points2.real, points1.real) + ducc0.misc.vdot(points2.imag, points1.imag)
        assert_allclose(v1, v2, rtol=1e-9)

@pmp('spin', (0, 1, 2))
@pmp('nthreads', (1, 4))
@pmp('nside', (32, ))
@pmp('lmmax', ((10, 8), (100,100)))
def test_synthesis_general(lmmax, nside, spin, nthreads):
    rng = np.random.default_rng(48)

    lmax, mmax = lmmax
    epsilon = 1e-7
    ncomp = 1 if spin == 0 else 2
    slm = random_alm(lmax, mmax, spin, ncomp, rng)
    base = ducc0.healpix.Healpix_Base(nside, "RING")
    geom = base.sht_info()
    loc = base.pix2ang(pix=np.arange(base.npix()), nthreads=nthreads)
    v1 = ducc0.sht.synthesis_general(lmax=lmax, mmax=mmax, alm=slm, loc=loc, spin=spin, epsilon=epsilon, nthreads=nthreads)
    v2 = ducc0.sht.synthesis(alm=slm, lmax=lmax, mmax=mmax, spin=spin, nthreads=nthreads, **geom)
    assert_allclose(ducc0.misc.l2error(v1,v2), 0, atol=epsilon)
